(* Modula-2 Library    -  UNIX SUN        -     HR 4/89 *)
(* (c) Universitaet Ulm, Sektion Informatik, D-7900 Ulm *)
IMPLEMENTATION MODULE TermCap;

   FROM Clock IMPORT UnitsPerSecond, RealTime;
   FROM Environment IMPORT GetEnv;
   FROM Strings IMPORT StrCat, StrLen, StrCmp, StrCpy;
   FROM StrToNum IMPORT StrToInt;
   FROM StdIO IMPORT Fopen, Fclose, read, FILE, Fgetc, Ftell, Fseek;
   FROM SYSTEM IMPORT ADDRESS, ADR;
   IMPORT Conversions, Strings, SysPanic, SYSTEM;

      TYPE
	 Tagtyp = ( bool, num, string );
	 Termfield = RECORD	(* database is not compiled *)
			sname : ARRAY[0..1] OF CHAR;
			CASE tag : Tagtyp OF
			  bool: Bool : BOOLEAN|
			  num: Number: INTEGER|
			  string: Text : String
			END(*CASE*)
		     END;
      CONST
	 anzstrings = 106;
	 anzbools = 26;
	 anzints =11;
	 lines = anzstrings + anzints + anzbools;

      VAR
	 table: ARRAY[0..lines-1] OF Termfield;

   MODULE Interpreter;

      FROM Conversions IMPORT ConvertInteger;
      FROM Strings IMPORT StrCat;
      FROM SysPanic IMPORT Panic;
      FROM SYSTEM IMPORT ADDRESS, BYTE, ADR;

      EXPORT Tparm, Tparm1, Tparm2, Tparm3, Tparm4, Tparm9;

      CONST
	 maxarg = 10;
      TYPE
	 ArgList = ARRAY[0..maxarg-1] OF ADDRESS;
      VAR
	 args: ArgList;     (* stack of arguments *)
	 arghigh: ARRAY [0..maxarg-1] OF CARDINAL; (* HIGH(args[]) *)
	 argi: CARDINAL;    (* index of next arg to be pushed *)

      PROCEDURE Abort(msg: ARRAY OF CHAR);
	 VAR panicmsg: ARRAY [0..79] OF CHAR;
      BEGIN
	 panicmsg := "TermCap.Tparm: ";
	 StrCat(panicmsg, msg);
	 Panic(panicmsg);
      END Abort;

      PROCEDURE PushArg(VAR a: ARRAY OF BYTE);
	 (* push next arg of Tparm? onto stack 'args' *)
      BEGIN
	 (* Assert(argi < maxarg); *)
	 args[argi] := ADR(a); arghigh[argi] := HIGH(a); INC(argi);
      END PushArg;

      PROCEDURE Print(VAR out: ARRAY OF CHAR; fmt: ARRAY OF CHAR);
	 TYPE
	    CharSet = SET OF CHAR;
	    VarName = ['a'..'z'];
	 CONST
	    Digits = CharSet{'0'..'9'};
	    VarNames = CharSet{MIN(VarName)..MAX(VarName)};
	    bitsperword = 32; (* size of BITSET in bits *)
	 VAR
	    index: CARDINAL;
	    outindex: CARDINAL;
	    zeropad: BOOLEAN;
	    padcnt: CARDINAL;
	    vars: ARRAY VarName OF INTEGER;
	    minus: BOOLEAN;
	    op: INTEGER; (* one operand *)
	    alev: CARDINAL;
	    iph : POINTER TO INTEGER;

	 MODULE Stack;

	    IMPORT Abort;

	    EXPORT Push, Pop;

	    CONST
	       stacksize = 10;
	    VAR
	       stack: ARRAY [0..stacksize-1] OF INTEGER;
	       top: CARDINAL;

	    PROCEDURE Push(arg: INTEGER);
	    BEGIN
	       IF top <= HIGH(stack) THEN
		  stack[top] := arg;
		  INC(top);
	       ELSE
		  Abort("stack overflow");
	       END;
	    END Push;

	    PROCEDURE Pop() : INTEGER;
	    BEGIN
	       IF top > 0 THEN
		  DEC(top);
		  RETURN stack[top]
	       ELSE
		  Abort("stack underflow");
	       END;
	    END Pop;

	 BEGIN
	    top := 0;
	 END Stack;

	 PROCEDURE Write(ch: CHAR);
	 BEGIN
	    IF outindex <= HIGH(out) THEN
	       out[outindex] := ch;
	       INC(outindex);
	    ELSE
	       Abort("overflow of output buffer");
	    END;
	 END Write;

	 PROCEDURE WriteString(s: ARRAY OF CHAR);
	    VAR index: CARDINAL;
	 BEGIN
	    index := 0;
	    WHILE (index <= HIGH(s)) AND (s[index] <> 0C) DO
	       Write(s[index]);
	       INC(index);
	    END;
	 END WriteString;

	 PROCEDURE Next;
	 BEGIN
	    IF index < HIGH(fmt) THEN
	       INC(index);
	    ELSE
	       Abort("unexpected end in format string");
	    END;
	 END Next;

	 PROCEDURE PopInteger;
	    VAR
	       field: ARRAY [0..11] OF CHAR;
	       index: CARDINAL;
	 BEGIN
	    ConvertInteger(Pop(), padcnt, field);
	    IF zeropad THEN
	       index := 0;
	       WHILE (index <= HIGH(field)) AND (field[index] = ' ') DO
		  field[index] := '0';
		  INC(index);
	       END;
	    END;
	    WriteString(field);
	 END PopInteger;

	 PROCEDURE PopChar;
	    CONST
	       minchar = VAL(INTEGER, ORD(MIN(CHAR)));
	       maxchar = VAL(INTEGER, ORD(MAX(CHAR)));
	    VAR
	       char: INTEGER;
	 BEGIN
	    char := Pop();
	    IF (char >= minchar) AND (char <= maxchar) THEN
	       Write(CHR(char));
	    ELSE
	       Abort("popped value of %c not in [0C..377C]");
	    END;
	 END PopChar;

	 PROCEDURE PushParm(i: CARDINAL);
	    VAR
	       ip: POINTER TO INTEGER;
	 BEGIN
	    IF (i = 0) OR (i > argi) THEN
	       Abort("bad parameter number");
	    END;
	    DEC(i);
	    ip := args[i];
	    Push(ip^);
	 END PushParm;

	 PROCEDURE Increment;
	    VAR ip: POINTER TO INTEGER;
	 BEGIN
	    IF argi < 1 THEN
	       Abort("%i: not enough parameters");
	    ELSIF argi = 1 THEN
	       ip := args[0]; INC(ip^);
	    ELSE
	       ip := args[0]; INC(ip^);
	       ip := args[1]; INC(ip^);
	    END;
	 END Increment;

	 PROCEDURE Skip(cset: CharSet);
	    VAR level: CARDINAL;
	 BEGIN
	    level := 0; Next;
	    LOOP
	       IF fmt[index] = '%' THEN
		  Next;
		  IF (level = 0) AND (fmt[index] IN cset) THEN
		     RETURN
		  END;
		  CASE fmt[index] OF
		  | '?': INC(level);
		  | ';': IF level > 0 THEN
			    DEC(level);
			 ELSE
			    Abort("%; without %?");
			 END;
		  ELSE
		  END;
	       END;
	       Next;
	    END;
	 END Skip;

      BEGIN (* Print *)
	 index := 0; outindex := 0; alev := 1;
	 WHILE (index <= HIGH(fmt)) AND (fmt[index] <> 0C) DO
	    IF fmt[index] = '%' THEN
	       Next;
	       zeropad := FALSE;
		padcnt := 0;
	       CASE fmt[index] OF
	       | 'd': PushParm(alev); INC(alev);
		      PopInteger;
	       | '2': padcnt := 2; PopInteger;
	       | '3': padcnt := 3; PopInteger;
	       | '.': PushParm(alev); INC(alev); PopChar;
	       | '+': Next;
		      args[alev -1] := args[alev -1] + ORD(fmt[index]);
		      PushParm(alev); INC(alev); PopChar;
	       | '>': Next;
		      IF args[alev -1] > ORD(fmt[index]) THEN
			  Next;
			  args[alev -1] := args[alev -1] + ORD(fmt[index]);
		      END;
	       | 'r': iph := args[alev];
			args[alev -1] := args[alev];
			args[alev] := iph;
	       | 'i': Increment;
	       | '%': Write("%");
	       | 'n': args[alev - 1] := CARDINAL( BITSET(args[alev -1]) /
					BITSET(0140));
		      args[alev] := CARDINAL( BITSET(args[alev])/
					BITSET(0140));
	       | 'B': args[alev -1] := (16 * (args[alev -1] DIV 10)) +							(args[alev -1 ] MOD 10 );
	       | 'D': args[alev -1] := (args[alev -1] - 2*(args[alev -1] 									MOD 16));
	       ELSE
		  Abort("illegal letter following %");
	       END;
	    ELSE
	       Write(fmt[index]);
	    END;
	    INC(index);
	 END;
	 IF outindex <= HIGH(out) THEN
	    out[outindex] := 0C;
	 END;
      END Print;

      PROCEDURE Tparm(VAR out: ARRAY OF CHAR; fmt: ARRAY OF CHAR);
      BEGIN
	 argi := 0;
	 Print(out, fmt);
      END Tparm;

      PROCEDURE Tparm1(VAR out: ARRAY OF CHAR; fmt: ARRAY OF CHAR;
		       arg1: ARRAY OF BYTE);
      BEGIN
	 argi := 0;
	 PushArg(arg1);
	 Print(out, fmt);
      END Tparm1;

      PROCEDURE Tparm2(VAR out: ARRAY OF CHAR; fmt: ARRAY OF CHAR;
		       arg1: ARRAY OF BYTE;
		       arg2: ARRAY OF BYTE);
      BEGIN
	 argi := 0;
	 PushArg(arg1);
	 PushArg(arg2);
	 Print(out, fmt);
      END Tparm2;

      PROCEDURE Tparm3(VAR out: ARRAY OF CHAR; fmt: ARRAY OF CHAR;
		       arg1: ARRAY OF BYTE;
		       arg2: ARRAY OF BYTE;
		       arg3: ARRAY OF BYTE);
      BEGIN
	 argi := 0;
	 PushArg(arg1);
	 PushArg(arg2);
	 PushArg(arg3);
	 Print(out, fmt);
      END Tparm3;

      PROCEDURE Tparm4(VAR out: ARRAY OF CHAR; fmt: ARRAY OF CHAR;
		       arg1: ARRAY OF BYTE;
		       arg2: ARRAY OF BYTE;
		       arg3: ARRAY OF BYTE;
		       arg4: ARRAY OF BYTE);
      BEGIN
	 argi := 0;
	 PushArg(arg1);
	 PushArg(arg2);
	 PushArg(arg3);
	 PushArg(arg4);
	 Print(out, fmt);
      END Tparm4;

      PROCEDURE Tparm9(VAR out: ARRAY OF CHAR; fmt: ARRAY OF CHAR;
		       arg1: ARRAY OF BYTE;
		       arg2: ARRAY OF BYTE;
		       arg3: ARRAY OF BYTE;
		       arg4: ARRAY OF BYTE;
		       arg5: ARRAY OF BYTE;
		       arg6: ARRAY OF BYTE;
		       arg7: ARRAY OF BYTE;
		       arg8: ARRAY OF BYTE;
		       arg9: ARRAY OF BYTE);
      BEGIN
	 argi := 0;
	 PushArg(arg1);
	 PushArg(arg2);
	 PushArg(arg3);
	 PushArg(arg4);
	 PushArg(arg5);
	 PushArg(arg6);
	 PushArg(arg7);
	 PushArg(arg8);
	 PushArg(arg9);
	 Print(out, fmt);
      END Tparm9;

   END Interpreter;

   PROCEDURE Delay(delay: CARDINAL);
      VAR
	 units: CARDINAL;
	 time: CARDINAL;
   BEGIN
      (* at least one system call *)
      time := RealTime((* reset = *) TRUE);
      units := delay * UnitsPerSecond DIV 10000;
      IF units > 0 THEN
	 (* every RealTime call causes a system call *)
	 (* so this busy-loop is fair to other processes *)
	 REPEAT
	    time := RealTime((* reset = *) FALSE);
	 UNTIL time > units;
      END;
   END Delay;

   PROCEDURE Tputs(str: ARRAY OF CHAR; affcnt: CARDINAL;
		   outc: OutProc);
   BEGIN
      TputsDelay(str, affcnt, outc, Delay);
   END Tputs;

   PROCEDURE TputsDelay(str: ARRAY OF CHAR; affcnt: CARDINAL;
			outc: OutProc; Delay: DelayProc);
	TYPE
	    CharSet = SET OF CHAR;
	CONST
	    Digits = CharSet{'0'..'9'};
        VAR
	    index: CARDINAL;
	    delay: CARDINAL;	(* delay in tenth of milliseconds *)

      PROCEDURE Pad;
	 VAR
	    start: CARDINAL;
      BEGIN
	 (* convert delay:  {digit} ["." {digit}] ["*"]  *)
	 start := index;
	 delay := 0;
	 WHILE (index <= HIGH(str)) AND (str[index] IN Digits) DO
	    delay := 10 * delay + ORD(str[index]) - ORD('0');
	    INC(index);
	 END;
	 delay := delay * 10;
	 IF (index <= HIGH(str)) AND (str[index] = '.') THEN
	    INC(index);
	    IF (index <= HIGH(str)) AND (str[index] IN Digits) THEN
	       INC(delay, ORD(str[index]) - ORD('0'));
	       INC(index);
	       WHILE (index <= HIGH(str)) AND (str[index] IN Digits) DO
		  INC(index);
	       END;
	    END;
	 END;
	 IF (index <= HIGH(str)) AND (str[index] = '*') THEN
	    INC(index);
	    delay := delay * affcnt;
	 END;

      END Pad;

   BEGIN
      index := 0;
      delay := 0;
      IF str[index] IN Digits THEN
	 Pad;
      ELSE
      	 WHILE (index <= HIGH(str)) AND (str[index] <> 0C) DO
	    outc(str[index]);
	    INC(index);
	 END;
      END;
      IF delay = 0 THEN
         RETURN
      END;
      Delay(delay);
   END TputsDelay;

   PROCEDURE InitTable( );

BEGIN
	table[0].sname := "MT";
	table[0].tag :=  bool;
	table[0].Bool := FALSE;
	table[1].sname := "am";
	table[1].tag :=  bool;
	table[1].Bool := FALSE;
	table[2].sname := "bs";
	table[2].tag :=  bool;
	table[2].Bool := FALSE;
	table[3].sname := "bw";
	table[3].tag :=  bool;
	table[3].Bool := FALSE;
	table[4].sname := "da";
	table[4].tag :=  bool;
	table[4].Bool := FALSE;
	table[5].sname := "db";
	table[5].tag :=  bool;
	table[5].Bool := FALSE;
	table[6].sname := "eo";
	table[6].tag :=  bool;
	table[6].Bool := FALSE;
	table[7].sname := "es";
	table[7].tag :=  bool;
	table[7].Bool := FALSE;
	table[8].sname := "hc";
	table[8].tag :=  bool;
	table[8].Bool := FALSE;
	table[9].sname := "hs";
	table[9].tag :=  bool;
	table[9].Bool := FALSE;
	table[10].sname := "in";
	table[10].tag :=  bool;
	table[10].Bool := FALSE;
	table[11].sname := "km";
	table[11].tag :=  bool;
	table[11].Bool := FALSE;
	table[12].sname := "mi";
	table[12].tag :=  bool;
	table[12].Bool := FALSE;
	table[13].sname := "ms";
	table[13].tag :=  bool;
	table[13].Bool := FALSE;
	table[14].sname := "nc";
	table[14].tag :=  bool;
	table[14].Bool := FALSE;
	table[15].sname := "ns";
	table[15].tag :=  bool;
	table[15].Bool := FALSE;
	table[16].sname := "os";
	table[16].tag :=  bool;
	table[16].Bool := FALSE;
	table[17].sname := "pl";
	table[17].tag :=  bool;
	table[17].Bool := FALSE;
	table[18].sname := "pt";
	table[18].tag :=  bool;
	table[18].Bool := FALSE;
	table[19].sname := "ul";
	table[19].tag :=  bool;
	table[19].Bool := FALSE;
	table[20].sname := "xb";
	table[20].tag :=  bool;
	table[20].Bool := FALSE;
	table[21].sname := "xn";
	table[21].tag :=  bool;
	table[21].Bool := FALSE;
	table[22].sname := "xo";
	table[22].tag :=  bool;
	table[22].Bool := FALSE;
	table[23].sname := "xr";
	table[23].tag :=  bool;
	table[23].Bool := FALSE;
	table[24].sname := "xs";
	table[24].tag :=  bool;
	table[24].Bool := FALSE;
	table[25].sname := "xt";
	table[25].tag :=  bool;
	table[25].Bool := FALSE;
	table[26].sname := "co";
	table[26].tag :=  num;
	table[26].Number := -1;
	table[27].sname := "dC";
	table[27].tag :=  num;
	table[27].Number := -1;
	table[28].sname := "dF";
	table[28].tag :=  num;
	table[28].Number := -1;
	table[29].sname := "dN";
	table[29].tag :=  num;
	table[29].Number := -1;
	table[30].sname := "dT";
	table[30].tag :=  num;
	table[30].Number := -1;
	table[31].sname := "kn";
	table[31].tag :=  num;
	table[31].Number := -1;
	table[32].sname := "li";
	table[32].tag :=  num;
	table[32].Number := -1;
	table[33].sname := "pb";
	table[33].tag :=  num;
	table[33].Number := -1;
	table[34].sname := "sg";
	table[34].tag :=  num;
	table[34].Number := -1;
	table[35].sname := "ug";
	table[35].tag :=  num;
	table[35].Number := -1;
	table[36].sname := "vt";
	table[36].tag :=  num;
	table[36].Number := -1;
	table[37].sname := "AL";
	table[37].tag :=  string;
	table[37].Text[0] := 0C;
	table[38].sname := "CC";
	table[38].tag :=  string;
	table[38].Text[0] := 0C;
	table[39].sname := "DC";
	table[39].tag :=  string;
	table[39].Text[0] := 0C;
	table[40].sname := "DL";
	table[40].tag :=  string;
	table[40].Text[0] := 0C;
	table[41].sname := "DO";
	table[41].tag :=  string;
	table[41].Text[0] := 0C;
	table[42].sname := "IC";
	table[42].tag :=  string;
	table[42].Text[0] := 0C;
	table[43].sname := "LE";
	table[43].tag :=  string;
	table[43].Text[0] := 0C;
	table[44].sname := "RI";
	table[44].tag :=  string;
	table[44].Text[0] := 0C;
	table[45].sname := "UP";
	table[45].tag :=  string;
	table[45].Text[0] := 0C;
	table[46].sname := "ae";
	table[46].tag :=  string;
	table[46].Text[0] := 0C;
	table[47].sname := "al";
	table[47].tag :=  string;
	table[47].Text[0] := 0C;
	table[48].sname := "as";
	table[48].tag :=  string;
	table[48].Text[0] := 0C;
	table[49].sname := "bc";
	table[49].tag :=  string;
	table[49].Text[0] := 0C;
	table[50].sname := "bl";
	table[50].tag :=  string;
	table[50].Text[0] := 0C;
	table[51].sname := "bt";
	table[51].tag :=  string;
	table[51].Text[0] := 0C;
	table[52].sname := "cd";
	table[52].tag :=  string;
	table[52].Text[0] := 0C;
	table[53].sname := "ce";
	table[53].tag :=  string;
	table[53].Text[0] := 0C;
	table[54].sname := "ch";
	table[54].tag :=  string;
	table[54].Text[0] := 0C;
	table[55].sname := "cl";
	table[55].tag :=  string;
	table[55].Text[0] := 0C;
	table[56].sname := "cm";
	table[56].tag :=  string;
	table[56].Text[0] := 0C;
	table[57].sname := "cr";
	table[57].tag :=  string;
	table[57].Text[0] := 0C;
	table[58].sname := "cs";
	table[58].tag :=  string;
	table[58].Text[0] := 0C;
	table[59].sname := "ct";
	table[59].tag :=  string;
	table[59].Text[0] := 0C;
	table[60].sname := "cv";
	table[60].tag :=  string;
	table[60].Text[0] := 0C;
	table[61].sname := "dc";
	table[61].tag :=  string;
	table[61].Text[0] := 0C;
	table[62].sname := "dl";
	table[62].tag :=  string;
	table[62].Text[0] := 0C;
	table[63].sname := "dm";
	table[63].tag :=  string;
	table[63].Text[0] := 0C;
	table[64].sname := "do";
	table[64].tag :=  string;
	table[64].Text[0] := 0C;
	table[65].sname := "ds";
	table[65].tag :=  string;
	table[65].Text[0] := 0C;
	table[66].sname := "ed";
	table[66].tag :=  string;
	table[66].Text[0] := 0C;
	table[67].sname := "ei";
	table[67].tag :=  string;
	table[67].Text[0] := 0C;
	table[68].sname := "ff";
	table[68].tag :=  string;
	table[68].Text[0] := 0C;
	table[69].sname := "fs";
	table[69].tag :=  string;
	table[69].Text[0] := 0C;
	table[70].sname := "hd";
	table[70].tag :=  string;
	table[70].Text[0] := 0C;
	table[71].sname := "ho";
	table[71].tag :=  string;
	table[71].Text[0] := 0C;
	table[72].sname := "hu";
	table[72].tag :=  string;
	table[72].Text[0] := 0C;
	table[73].sname := "i2";
	table[73].tag :=  string;
	table[73].Text[0] := 0C;
	table[74].sname := "ic";
	table[74].tag :=  string;
	table[74].Text[0] := 0C;
	table[75].sname := "if";
	table[75].tag :=  string;
	table[75].Text[0] := 0C;
	table[76].sname := "im";
	table[76].tag :=  string;
	table[76].Text[0] := 0C;
	table[77].sname := "ip";
	table[77].tag :=  string;
	table[77].Text[0] := 0C;
	table[78].sname := "is";
	table[78].tag :=  string;
	table[78].Text[0] := 0C;
	table[79].sname := "k0";
	table[79].tag :=  string;
	table[79].Text[0] := 0C;
	table[80].sname := "k1";
	table[80].tag :=  string;
	table[80].Text[0] := 0C;
	table[81].sname := "k2";
	table[81].tag :=  string;
	table[81].Text[0] := 0C;
	table[82].sname := "k3";
	table[82].tag :=  string;
	table[82].Text[0] := 0C;
	table[83].sname := "k4";
	table[83].tag :=  string;
	table[83].Text[0] := 0C;
	table[84].sname := "k5";
	table[84].tag :=  string;
	table[84].Text[0] := 0C;
	table[85].sname := "k6";
	table[85].tag :=  string;
	table[85].Text[0] := 0C;
	table[86].sname := "k7";
	table[86].tag :=  string;
	table[86].Text[0] := 0C;
	table[87].sname := "k8";
	table[87].tag :=  string;
	table[87].Text[0] := 0C;
	table[88].sname := "k9";
	table[88].tag :=  string;
	table[88].Text[0] := 0C;
	table[89].sname := "kb";
	table[89].tag :=  string;
	table[89].Text[0] := 0C;
	table[90].sname := "kd";
	table[90].tag :=  string;
	table[90].Text[0] := 0C;
	table[91].sname := "ke";
	table[91].tag :=  string;
	table[91].Text[0] := 0C;
	table[92].sname := "kh";
	table[92].tag :=  string;
	table[92].Text[0] := 0C;
	table[93].sname := "kl";
	table[93].tag :=  string;
	table[93].Text[0] := 0C;
	table[94].sname := "ko";
	table[94].tag :=  string;
	table[94].Text[0] := 0C;
	table[95].sname := "kr";
	table[95].tag :=  string;
	table[95].Text[0] := 0C;
	table[96].sname := "ks";
	table[96].tag :=  string;
	table[96].Text[0] := 0C;
	table[97].sname := "ku";
	table[97].tag :=  string;
	table[97].Text[0] := 0C;
	table[98].sname := "l0";
	table[98].tag :=  string;
	table[98].Text[0] := 0C;
	table[99].sname := "l1";
	table[99].tag :=  string;
	table[99].Text[0] := 0C;
	table[100].sname := "l2";
	table[100].tag :=  string;
	table[100].Text[0] := 0C;
	table[101].sname := "l3";
	table[101].tag :=  string;
	table[101].Text[0] := 0C;
	table[102].sname := "l4";
	table[102].tag :=  string;
	table[102].Text[0] := 0C;
	table[103].sname := "l5";
	table[103].tag :=  string;
	table[103].Text[0] := 0C;
	table[104].sname := "l6";
	table[104].tag :=  string;
	table[104].Text[0] := 0C;
	table[105].sname := "l7";
	table[105].tag :=  string;
	table[105].Text[0] := 0C;
	table[106].sname := "l8";
	table[106].tag :=  string;
	table[106].Text[0] := 0C;
	table[107].sname := "le";
	table[107].tag :=  string;
	table[107].Text[0] := 0C;
	table[108].sname := "ll";
	table[108].tag :=  string;
	table[108].Text[0] := 0C;
	table[109].sname := "ma";
	table[109].tag :=  string;
	table[109].Text[0] := 0C;
	table[110].sname := "mb";
	table[110].tag :=  string;
	table[110].Text[0] := 0C;
	table[111].sname := "md";
	table[111].tag :=  string;
	table[111].Text[0] := 0C;
	table[112].sname := "me";
	table[112].tag :=  string;
	table[112].Text[0] := 0C;
	table[113].sname := "mh";
	table[113].tag :=  string;
	table[113].Text[0] := 0C;
	table[114].sname := "mk";
	table[114].tag :=  string;
	table[114].Text[0] := 0C;
	table[115].sname := "mp";
	table[115].tag :=  string;
	table[115].Text[0] := 0C;
	table[116].sname := "mr";
	table[116].tag :=  string;
	table[116].Text[0] := 0C;
	table[117].sname := "nd";
	table[117].tag :=  string;
	table[117].Text[0] := 0C;
	table[118].sname := "nl";
	table[118].tag :=  string;
	table[118].Text[0] := 0C;
	table[119].sname := "nw";
	table[119].tag :=  string;
	table[119].Text[0] := 0C;
	table[120].sname := "pc";
	table[120].tag :=  string;
	table[120].Text[0] := 0C;
	table[121].sname := "rc";
	table[121].tag :=  string;
	table[121].Text[0] := 0C;
	table[122].sname := "rf";
	table[122].tag :=  string;
	table[122].Text[0] := 0C;
	table[123].sname := "ri";
	table[123].tag :=  string;
	table[123].Text[0] := 0C;
	table[124].sname := "rs";
	table[124].tag :=  string;
	table[124].Text[0] := 0C;
	table[125].sname := "sc";
	table[125].tag :=  string;
	table[125].Text[0] := 0C;
	table[126].sname := "se";
	table[126].tag :=  string;
	table[126].Text[0] := 0C;
	table[127].sname := "sf";
	table[127].tag :=  string;
	table[127].Text[0] := 0C;
	table[128].sname := "so";
	table[128].tag :=  string;
	table[128].Text[0] := 0C;
	table[129].sname := "sr";
	table[129].tag :=  string;
	table[129].Text[0] := 0C;
	table[130].sname := "st";
	table[130].tag :=  string;
	table[130].Text[0] := 0C;
	table[131].sname := "ta";
	table[131].tag :=  string;
	table[131].Text[0] := 0C;
	table[132].sname := "tc";
	table[132].tag :=  string;
	table[132].Text[0] := 0C;
	table[133].sname := "te";
	table[133].tag :=  string;
	table[133].Text[0] := 0C;
	table[134].sname := "ti";
	table[134].tag :=  string;
	table[134].Text[0] := 0C;
	table[135].sname := "ts";
	table[135].tag :=  string;
	table[135].Text[0] := 0C;
	table[136].sname := "uc";
	table[136].tag :=  string;
	table[136].Text[0] := 0C;
	table[137].sname := "ue";
	table[137].tag :=  string;
	table[137].Text[0] := 0C;
	table[138].sname := "up";
	table[138].tag :=  string;
	table[138].Text[0] := 0C;
	table[139].sname := "us";
	table[139].tag :=  string;
	table[139].Text[0] := 0C;
	table[140].sname := "vb";
	table[140].tag :=  string;
	table[140].Text[0] := 0C;
	table[141].sname := "ve";
	table[141].tag :=  string;
	table[141].Text[0] := 0C;
	table[142].sname := "vs";
	table[142].tag :=  string;
	table[142].Text[0] := 0C;
   END InitTable;

   PROCEDURE SetupTerm(termname: ARRAY OF CHAR; VAR tinfo: Term) : BOOLEAN;
      CONST
	 defaultfile = "/usr/etc/termcap";
	 caplen = 639;
      TYPE
	 OFF = INTEGER;			(* offset type as used by StdIO.Ftell *)
	 Short = INTEGER;
      VAR
	 ok: BOOLEAN;			(* result of GetEnv *)
	 tcap: ARRAY [0..caplen] OF CHAR;	(* capabilities for terminal*)
	 tname: ARRAY [0..31] OF CHAR;	(* terminal name *)
	 dir: ARRAY[0..2] OF CHAR;	(* first letter of terminal name *)
	 tfp: FILE;			(* terminfo file *)
	 done: BOOLEAN;		(* set to FALSE if anything goes wrong *)
	 captable: ARRAY[0..lines-1] OF Termfield;
	 index : CARDINAL;

(* find the right termcap entry in termcap *)
PROCEDURE FindCap( fp : FILE; VAR tcap : ARRAY OF CHAR) : BOOLEAN;
TYPE
	ZUSTAND = ( Start, Comment, Lese1, Lese2, Capability );
VAR 
	zustand : ZUSTAND;
  	i,j : INTEGER;
	ch : CHAR;
	buf : ARRAY[0..300] OF CHAR;

   PROCEDURE TestTerm(buf: ARRAY OF CHAR; term : ARRAY OF CHAR) : BOOLEAN;

	BEGIN
	IF ( StrCmp(buf, term) = 0) THEN
	   RETURN TRUE
	ELSE
	   RETURN FALSE
	END;
   END TestTerm;

(* read the termcap entry into tcap *)

   PROCEDURE GetCap( VAR tcap : ARRAY OF CHAR; buf : ARRAY OF CHAR; fp: FILE);
   VAR
      lese : BOOLEAN;
      i : CARDINAL;
      ch : CHAR;
   BEGIN
      lese := TRUE;
      StrCat(tcap , buf);
      StrCat(tcap , "|");
      i := StrLen(tcap) ;
      WHILE lese DO
	 WHILE Fgetc(ch, fp) AND ( ch <> 12C) DO
	    tcap[i] := ch;
	    INC(i);
	 END;
	 IF tcap[i-1] = ":" THEN
	    lese := FALSE;
	 ELSE
	    tcap[i] := ch;
	    INC(i);
	 END;
      END;
   END GetCap;
BEGIN
  zustand := Start;
  WHILE Fgetc(ch, fp ) DO
      CASE zustand OF
      | Start: i := 0;
		IF ( ch = "#" ) THEN

		   zustand := Comment;
		ELSE 
		   tcap[i] := ch;
		   i := i+1;
		   zustand := Lese1;
		END;
      | Comment:  WHILE (ch <> 12C) DO
		 IF NOT Fgetc(ch , fp) THEN
		    ch := 12C;
		  END;
		 END;
		 zustand := Start;
      | Lese1:   WHILE ( ch <> "|") DO
			tcap[i] := ch;
			i := i +1;
			IF NOT Fgetc(ch,fp) THEN
			   ch :="|";
			END;
		 END;
		 tcap[i] := ch;
		 tcap[i+1] := 0C;
		 zustand :=Lese2;
      | Lese2:  j := 0;
		REPEAT
		   buf[j] := ch;
		   INC(j);
		UNTIL NOT( Fgetc(ch, fp) AND ( ch <> "|") AND ( ch <> ":")) ;
		buf[j] := 0C;
		IF ch = "|" THEN
		   IF TestTerm("sun", buf) THEN
		      GetCap(tcap, buf, fp);
		      RETURN TRUE
		   END;
		   zustand := Lese2;
		ELSE IF ch = ":" THEN
		   zustand := Capability;
		END;
		END;
      | Capability: WHILE ( ch <> ":") DO
			IF Fgetc(ch, fp) THEN
			ELSE
			   ch := ":";
			END;
		     END;
			IF Fgetc(ch, fp) THEN
			   IF ch =  12C THEN
			      zustand := Start;
			   ELSE
			      zustand := Capability;
			   END;
			END;
      END;
   END;
   RETURN FALSE
		  
END FindCap;

(*set the capabilities of terminal tname *)
PROCEDURE ReadTcap() : BOOLEAN;
TYPE
	Zustand = ( start, capab, newl, readcap, zbool, znum, zstring);
VAR
	buf : ARRAY[0..31] OF CHAR;
	wert : INTEGER;
	caps : ARRAY[0..1] OF CHAR;
	zustand : Zustand;
	i,j : CARDINAL;
	typ : Tagtyp;

      PROCEDURE BinarySearch(search: ARRAY OF CHAR; begin, end : CARDINAL): 								INTEGER;
      VAR
	middle : INTEGER;
      BEGIN
      WHILE (begin < end) DO
	 middle := ( begin + end) DIV 2;
	 IF (StrCmp(search, captable[middle].sname) =0 ) THEN
	    RETURN middle
	 ELSE IF (StrCmp(search, captable[middle].sname) <0 ) THEN
	    end  := middle;
	 ELSE 
	    begin := middle;
	 END;
	 END;
      END;
      RETURN -1
      END BinarySearch;

      PROCEDURE InsertBool( scap: ARRAY OF CHAR ) : BOOLEAN;
      VAR
	found : INTEGER;
	anfang, ende : INTEGER;
      BEGIN
      anfang := 0; ende := anzbools -1;
      found := BinarySearch( scap, anfang, ende);
      IF (found >= 0) AND ( captable[found].tag = bool) THEN
	 StrCpy(captable[found].sname , scap);
	 RETURN TRUE
      ELSE
	 RETURN FALSE
      END;
      END InsertBool;

      PROCEDURE InsertNum(scap : ARRAY OF CHAR; value : INTEGER) : BOOLEAN;
      VAR
	found : INTEGER;
	anfang, ende : CARDINAL;
      BEGIN
      anfang := anzbools; ende := anfang + anzints -1;
      found := BinarySearch(scap, anfang, ende);
      IF (found > 0) AND ( captable[found].tag = num) THEN
	 StrCpy(captable[found].sname , scap);
	 captable[found].Number := value;
	 RETURN TRUE
      ELSE
	 RETURN FALSE
      END;
      END InsertNum;

      PROCEDURE InsertString(scap: ARRAY OF CHAR;text: ARRAY OF CHAR): BOOLEAN;
      VAR
	found : INTEGER;
	anfang,ende : CARDINAL;
      BEGIN
      anfang:= anzbools + anzints; ende := anfang + anzstrings -1;
      found := BinarySearch(scap, anfang, ende);
      IF (found > 0) AND ( captable[found].tag = string) THEN
	 StrCpy(captable[found].sname , scap);
	 StrCpy(captable[found].Text , text);
	 RETURN TRUE
      ELSE
	 RETURN FALSE
      END;
      END InsertString;
BEGIN
   zustand := start;
   i := 0;
   WHILE tcap[i] <> 0C DO
     CASE zustand OF
     | start: WHILE tcap[i] <> ":" DO
		INC(i);
	      END;
	      zustand := capab;
	      INC(i);
     | capab: IF tcap[i] = "\" THEN
		zustand := newl;
	     ELSE
		zustand := readcap;
	     END;
     | newl: WHILE tcap[i] <> ":" DO
		INC(i);
	    END;
	    INC(i);
	    zustand := readcap;
     | readcap: caps[0] := tcap[i];
	       INC(i);
	       caps[1] := tcap[i];
	       INC(i);
	       IF tcap[i] = ":" THEN
		  typ := bool;
		  zustand := zbool;
	       ELSE IF tcap[i] = "=" THEN
		  typ := string;
		  zustand := zstring
	       ELSE IF tcap[i] = "#" THEN
		  typ := num;
		  zustand := znum;
	       END;
	       END;
	       END;
	       INC(i);
     | zbool: IF NOT InsertBool(caps) THEN
		RETURN FALSE
	    END;
	    zustand := capab;
     | znum: j :=0;
	   WHILE tcap[i] <> ":" DO
	      buf[j] := tcap[i];
	      INC(i); INC(j);
	   END;
	   buf[j] := 0C;
	   IF NOT StrToInt( buf, wert) THEN
	       RETURN FALSE
	   END;
	   IF NOT InsertNum(caps, wert) THEN
	       RETURN FALSE
	    END;
	    INC(i);
	    zustand := capab;
     | zstring:   j := 0;
		WHILE tcap[i] <> ":" DO
		     IF tcap[i] = "^" THEN
			 INC(i);
			 buf[j] := CHR(ORD(tcap[i]) - ORD('A') +1);
		     ELSE IF (tcap[i] = '\') AND (tcap[i +1] = 'E') THEN
			 buf[j] := CHR(27);
			 INC(i);
		     ELSE
			 buf[j] := tcap[i];
		     END;
		     END;
		     INC(i); INC(j);
	     	 END;
		 buf[j] := 0C;
	   	 IF NOT InsertString(caps, buf) THEN
	        	 RETURN FALSE
	   	 END;
	     	 zustand := capab;
	     	 INC(i);
     END(*CASE*);
   END(*WHILE*);
   RETURN TRUE

END ReadTcap;
  
   (*give the capabilities to terminfo*)
   PROCEDURE InitTerm( VAR terminfo : Term );
   BEGIN
           terminfo.TMT := captable[0].Bool;
           terminfo.AutoRightMargin := captable[1].Bool;
           terminfo.Tbs := captable[2].Bool;
           terminfo.AutoLeftMargin := captable[3].Bool;
           terminfo.MemoryAbove := captable[4].Bool;
           terminfo.MemoryBelow := captable[5].Bool;
           terminfo.EraseOverstrike := captable[6].Bool;
           terminfo.Tes := captable[7].Bool;
           terminfo.HardCopy := captable[8].Bool;
           terminfo.Ths := captable[9].Bool;
           terminfo.InsertNullGlitch := captable[10].Bool;
           terminfo.Tkm := captable[11].Bool;
           terminfo.MoveInsertMode := captable[12].Bool;
           terminfo.Tms := captable[13].Bool;
           terminfo.Tnc := captable[14].Bool;
           terminfo.Tns := captable[15].Bool;
           terminfo.OverStrike := captable[16].Bool;
           terminfo.Tpl := captable[17].Bool;
           terminfo.Tpt := captable[18].Bool;
           terminfo.TransparentUnderline := captable[19].Bool;
           terminfo.BeehiveGlitch := captable[20].Bool;
           terminfo.EatNewlineGlitch := captable[21].Bool;
           terminfo.Txo := captable[22].Bool;
           terminfo.Txr := captable[23].Bool;
           terminfo.CeolStandoutGlitch := captable[24].Bool;
           terminfo.TelerayGlitch := captable[25].Bool;
           terminfo.Columns := captable[26].Number;
           terminfo.TdC := captable[27].Number;
           terminfo.TdF := captable[28].Number;
           terminfo.TdN := captable[29].Number;
           terminfo.TdT := captable[30].Number;
           terminfo.Tkn := captable[31].Number;
           terminfo.Lines := captable[32].Number;
           terminfo.Tpb := captable[33].Number;
           terminfo.MagicCookieGlitch := captable[34].Number;
           terminfo.Tug := captable[35].Number;
           terminfo.Tvt := captable[36].Number;
           terminfo.TAL := captable[37].Text;
           terminfo.CommandCharacter := captable[38].Text;
           terminfo.TDC := captable[39].Text;
           terminfo.TDL := captable[40].Text;
           terminfo.TDO := captable[41].Text;
           terminfo.TIC := captable[42].Text;
           terminfo.TLE := captable[43].Text;
           terminfo.TRI := captable[44].Text;
           terminfo.TUP := captable[45].Text;
           terminfo.ExitAltCharsetMode := captable[46].Text;
           terminfo.InsertLine := captable[47].Text;
           terminfo.EnterAltCharsetMode := captable[48].Text;
           terminfo.Tbc := captable[49].Text;
           terminfo.Tbl := captable[50].Text;
           terminfo.BackTab := captable[51].Text;
           terminfo.ClrEos := captable[52].Text;
           terminfo.ClrEol := captable[53].Text;
           terminfo.ColumnAddress := captable[54].Text;
           terminfo.ClearScreen := captable[55].Text;
           terminfo.CursorAddress := captable[56].Text;
           terminfo.CarriageReturn := captable[57].Text;
           terminfo.ChangeScrollRegion := captable[58].Text;
           terminfo.Tct := captable[59].Text;
           terminfo.RowAddress := captable[60].Text;
           terminfo.DeleteCharacter := captable[61].Text;
           terminfo.DeleteLine := captable[62].Text;
           terminfo.EnterDeleteMode := captable[63].Text;
           terminfo.CursorDown := captable[64].Text;
           terminfo.Tds := captable[65].Text;
           terminfo.ExitDeleteMode := captable[66].Text;
           terminfo.ExitInsertMode := captable[67].Text;
           terminfo.FormFeed := captable[68].Text;
           terminfo.Tfs := captable[69].Text;
           terminfo.DownHalfLine := captable[70].Text;
           terminfo.CursorHome := captable[71].Text;
           terminfo.UpHalfLine := captable[72].Text;
           terminfo.Ti2 := captable[73].Text;
           terminfo.InsertCharacter := captable[74].Text;
           terminfo.InitFile := captable[75].Text;
           terminfo.EnterInsertMode := captable[76].Text;
           terminfo.InsertPadding := captable[77].Text;
           terminfo.Tis := captable[78].Text;
           terminfo.KeyF0 := captable[79].Text;
           terminfo.KeyF1 := captable[80].Text;
           terminfo.KeyF2 := captable[81].Text;
           terminfo.KeyF3 := captable[82].Text;
           terminfo.KeyF4 := captable[83].Text;
           terminfo.KeyF5 := captable[84].Text;
           terminfo.KeyF6 := captable[85].Text;
           terminfo.KeyF7 := captable[86].Text;
           terminfo.KeyF8 := captable[87].Text;
           terminfo.KeyF9 := captable[88].Text;
           terminfo.KeyBackspace := captable[89].Text;
           terminfo.KeyDown := captable[90].Text;
           terminfo.KeypadLocal := captable[91].Text;
           terminfo.KeyHome := captable[92].Text;
           terminfo.KeyLeft := captable[93].Text;
           terminfo.Tko := captable[94].Text;
           terminfo.KeyRight := captable[95].Text;
           terminfo.KeypadXmit := captable[96].Text;
           terminfo.KeyUp := captable[97].Text;
           terminfo.LabF0 := captable[98].Text;
           terminfo.LabF1 := captable[99].Text;
           terminfo.LabF2 := captable[100].Text;
           terminfo.LabF3 := captable[101].Text;
           terminfo.LabF4 := captable[102].Text;
           terminfo.LabF5 := captable[103].Text;
           terminfo.LabF6 := captable[104].Text;
           terminfo.LabF7 := captable[105].Text;
           terminfo.LabF8 := captable[106].Text;
           terminfo.Tle := captable[107].Text;
           terminfo.CursorToLl := captable[108].Text;
           terminfo.Tma := captable[109].Text;
           terminfo.Tmb := captable[110].Text;
           terminfo.Tmd := captable[111].Text;
           terminfo.Tme := captable[112].Text;
           terminfo.Tmh := captable[113].Text;
           terminfo.Tmk := captable[114].Text;
           terminfo.Tmp := captable[115].Text;
           terminfo.Tmr := captable[116].Text;
           terminfo.CursorRight := captable[117].Text;
           terminfo.Tnl := captable[118].Text;
           terminfo.Tnw := captable[119].Text;
           terminfo.PadChar := captable[120].Text;
           terminfo.Trc := captable[121].Text;
           terminfo.Trf := captable[122].Text;
           terminfo.Tri := captable[123].Text;
           terminfo.Trs := captable[124].Text;
           terminfo.Tsc := captable[125].Text;
           terminfo.ExitStandoutMode := captable[126].Text;
           terminfo.Tsf := captable[127].Text;
           terminfo.EnterStandoutMode := captable[128].Text;
           terminfo.ScrollReverse := captable[129].Text;
           terminfo.Tst := captable[130].Text;
           terminfo.Tab := captable[131].Text;
           terminfo.Ttc := captable[132].Text;
           terminfo.ExitCaMode := captable[133].Text;
           terminfo.EnterCaMode := captable[134].Text;
           terminfo.Tts := captable[135].Text;
           terminfo.UnderlineChar := captable[136].Text;
           terminfo.ExitUnderlineMode := captable[137].Text;
           terminfo.CursorUp := captable[138].Text;
           terminfo.EnterUnderlineMode := captable[139].Text;
           terminfo.FlashScreen := captable[140].Text;
           terminfo.CursorNormal := captable[141].Text;
           terminfo.CursorVisible := captable[142].Text;
   END InitTerm;

   BEGIN
      IF termname[0] = 0C THEN
	 (* terminal name given? *)
	 GetEnv("TERM", tname, ok);
	 IF NOT ok THEN RETURN FALSE END;
      ELSE
	 StrCpy(tname, termname);
      END;
      GetEnv("TERMCAP", tcap, ok);
      IF NOT ok THEN
      	IF NOT Fopen(tfp, defaultfile, read, (* buffered = *) TRUE) THEN
		 RETURN FALSE
      	END;
	IF NOT FindCap(tfp, tcap) THEN
	   RETURN FALSE
	END(*IF*);
      END;
      FOR index := 0 TO lines -1 DO
         captable[index] := table[index];
      END(*FOR*);
      IF NOT ReadTcap() THEN
	   RETURN FALSE
      END;
      InitTerm(tinfo);
      RETURN TRUE
   END SetupTerm;
BEGIN
   InitTable( );

END TermCap.
